#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import ast
import inspect
import os
import re
from importlib.util import module_from_spec, spec_from_file_location
from inspect import (
    Parameter,
    getdoc,
    getmembers,
    getsourcefile,
    getsourcelines,
    isbuiltin,
    isclass,
    isfunction,
    signature,
)
from os.path import exists
from pathlib import Path
from textwrap import dedent
from types import LambdaType
from typing import Any, Callable, ClassVar, Dict, Generic, List, Optional, Set, TypeVar, Union

from pydantic import Field, constr, root_validator, validator
from pydantic.fields import Undefined
from typing_extensions import ParamSpec

from dbgen.configuration import config
from dbgen.core.base import Base
from dbgen.exceptions import DBgenInternalError, InvalidArgument
from dbgen.utils.misc import reserved_words


class Import(Base):
    """
    Representation of an Python import line.

    Examples:

    --> from libname import unaliased, things, aliased as Thing
        Import('libname',         <---- (DO NOT make this a keyword argument!!!)
                'unaliased',
                'things',
                aliased = 'Thing')

    --> import numpy as np
        Import('numpy', alias = 'np')

    Unsupported edge case: trying to import some variable literally named
                            "alias" using an alias
    """

    lib: constr(min_length=1)  # type: ignore
    unaliased_imports: Union[str, List[str]] = Field(default_factory=list)
    lib_alias: Optional[str] = None
    aliased_imports: Dict[str, str] = Field(default_factory=dict)
    _reserved_words: ClassVar[Set[str]] = reserved_words

    def __init__(self, lib, unaliased_imports=None, lib_alias=None, aliased_imports=None):
        unaliased_imports = unaliased_imports or []
        aliased_imports = aliased_imports or {}
        super().__init__(
            lib=lib,
            lib_alias=lib_alias,
            unaliased_imports=unaliased_imports,
            aliased_imports=aliased_imports,
        )

    @root_validator(pre=True)
    def check_lib_alias(cls, values):
        terms = values.get("unaliased_imports") or values.get("aliased_imports")
        lib_alias = values.get("lib_alias")
        lib = values.get("lib")
        err = f"Can't import {lib} as {lib_alias} AND import specific terms ({terms}) at once"
        assert not (lib_alias and terms), err
        return values

    @validator("unaliased_imports", pre=True)
    def singleton_list(cls, unaliased_imports):
        if isinstance(unaliased_imports, str):
            return [unaliased_imports]
        else:
            return unaliased_imports or []

    @validator("aliased_imports")
    def reserved_words(cls, aliased_imports):
        for alias in aliased_imports.values():
            assert alias not in cls._reserved_words, f"Reserved python word used as alias: {alias}"
        return aliased_imports

    @validator("lib_alias")
    def check_reserbed_lib_alias(cls, lib_alias):
        assert lib_alias not in cls._reserved_words, f"Reserved python word used as lib_alias: {lib_alias}"
        return lib_alias

    def __str__(self) -> str:
        if not (self.unaliased_imports or self.aliased_imports):
            alias = f"as {self.lib_alias}" if self.lib_alias else ""
            return f"import {self.lib} {alias}"
        else:
            aliased_imports = [f"{k} as {v}" for k, v in self.aliased_imports.items()]
            terms = list(self.unaliased_imports) + aliased_imports
            return f"from {self.lib} import {', '.join(terms)}"

    def __lt__(self, other: Any) -> bool:
        if type(self) == type(other):
            return repr(self) < repr(other)
        raise TypeError(f"'<' not supported between instances of '{type(self)}' and '{type(other)}'")


class Environment(Base):
    """
    Environment in which a python statement gets executed
    """

    imports: List[Import] = Field(default_factory=lambda: list())

    def __init__(self, imports=None):
        if isinstance(imports, Import):
            imports = [imports]
        imports = imports or []
        super().__init__(imports=imports)

    @validator("imports")
    def sort_imports(cls, imports):
        return sorted(imports)

    def __str__(self) -> str:
        return "\n".join(map(str, self.imports))

    def __add__(self, other: "Environment") -> "Environment":
        return Environment(imports=list(set(self.imports + other.imports)))

    # Public methods #

    @staticmethod
    def from_str(import_string: str) -> "Environment":
        """Parse a header"""
        imports = []
        for node in ast.iter_child_nodes(ast.parse(import_string)):
            if isinstance(node, ast.ImportFrom):
                lib = node.module
                assert lib, f"Empty lib found {lib}\n{import_string}"
                aliased_imports = {}
                unaliased_imports = []
                for module in node.names:
                    if module.asname:
                        aliased_imports[module.name] = module.asname
                    else:
                        unaliased_imports.append(module.name)
                imports.append(
                    Import(
                        lib=lib,
                        unaliased_imports=unaliased_imports,
                        aliased_imports=aliased_imports,
                    )
                )
            elif isinstance(node, ast.Import):
                assert len(node.names) == 1, f"Bad Import String! {import_string}"
                lib = node.names[0].name
                lib_alias = node.names[0].asname or ""
                imports.append(Import(lib=lib, lib_alias=lib_alias))
        return Environment(imports=imports)

    @classmethod
    def from_file(cls, pth: Path) -> "Environment":
        with open(pth) as f:
            return cls.from_str(f.read())


Input = ParamSpec('Input')
Output = TypeVar('Output')
FuncIn = ParamSpec('FuncIn')
FuncOut = TypeVar('FuncOut')


class Func(Base, Generic[FuncOut]):
    """
    A function that can be used during the DB generation process.
    """

    src: str
    name: str
    env: Environment
    _func: Optional[Callable]

    class Config:
        """Pydantic Config"""

        underscore_attrs_are_private = True

    def __init__(self, *, function: Optional[Callable] = None, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = function.__name__ if function else 'no_name'
        super().__init__(**kwargs)
        self._func = function

    def __str__(self) -> str:
        n = self.src.count("\n")
        s = "" if n == 1 else "s"
        return "<Func (%d line%s)>" % (n, s)

    def __call__(self, *args, **kwargs) -> FuncOut:
        if self._func is not None and self.path.exists():
            return self._func(*args, **kwargs)
        else:
            f = self._from_src()
            return f(*args, **kwargs)

    def __repr__(self) -> str:
        return self.name

    # Properties #

    @property
    def is_lam(self) -> bool:
        return self.src[:6] == "lambda"

    @property
    def doc(self) -> str:
        return getdoc(self._from_src()) or ""

    @property
    def sig(self) -> Any:
        return signature(self._from_src())

    @property
    def argnames(self) -> List[str]:
        return list(self.sig.parameters)

    @property
    def number_of_inputs(self) -> int:
        return len(self.argnames)

    @property
    def var_positional_keyword(self):
        return any(p.kind == Parameter.VAR_POSITIONAL for p in self.sig.parameters.values())

    @property
    def number_of_required_inputs(self) -> int:
        return len([k for k, v in self.sig.parameters.items() if v.default is v.empty])

    @property
    def notImp(self) -> bool:
        return "NotImplementedError" in self.src

    @property
    def output(self) -> Any:
        return self.sig.return_annotation

    @property
    def number_of_outputs(self) -> int:
        return 1

    # @property
    # def inTypes(self) -> List["DataType"]:
    #     return [
    #         DataType.get_datatype(x.annotation) for x in self.sig.parameters.values()
    #     ]

    @property
    def path(self) -> Path:
        return config.temp_dir / f"{self.name}_{self.hash}.py"

    def file(self) -> str:
        lam = "f = " if self.is_lam else ""
        return str(self.env) + "\n" + lam + self.src

    # Private methods #

    def _from_src(self, force: bool = False) -> Callable:
        """
        Execute source code to get a callable
        """

        if force or not exists(self.path):
            with open(self.path, "w") as t:
                t.write(self.file())

        if force or self._func is None:
            self._func = self.path_to_func(str(self.path))

        return self._func

    # # Public methods #
    def store_func(self, force: bool = False) -> None:
        """Load func from source code and store as attribute (better performance
        but object is no longer serializable / comparable for equality )
        """
        self._func = self._from_src(force)

    def set_func(self, value: Optional[Callable]) -> None:
        """Set the value of Func._func.

        Useful for removing the function object from the Func object for pickling
        """
        self._func = value

    @staticmethod
    def path_to_func(pth: str) -> Callable:

        try:
            spec = spec_from_file_location("random", pth)
            assert spec and spec.loader, "Spec or Spec.loader are broken"
            mod = module_from_spec(spec)
            spec.loader.exec_module(mod)  # type: ignore
            transforms = [o for o in getmembers(mod) if isfunction(o[1]) and getsourcefile(o[1]) == pth]
            assert len(transforms) == 1, "Bad input file %s has %d functions, not 1" % (
                pth,
                len(transforms),
            )
            return transforms[0][1]

        except Exception as e:
            if exists(pth):
                with open(pth) as f:
                    content = f.read()
            raise DBgenInternalError(
                f"Error while trying to load source code. You may be missing an import in your Transforms Env object. \nPath:{pth}\nFile Contents:\n--------\n{content}\n--------\nLoad Error: {e}"
            )


def get_short_lambda_source(lambda_func):
    """Return the source of a (short) lambda function.
    If it's impossible to obtain, returns None.
    """
    try:
        source_lines, _ = inspect.getsourcelines(lambda_func)
    except (OSError, TypeError):
        return None

    # skip `def`-ed functions and long lambdas
    if len(source_lines) != 1:
        return None

    source_text = os.linesep.join(source_lines).strip()

    # find the AST node of a lambda definition
    # so we can locate it in the source code
    try:
        source_ast = ast.parse(source_text)
    except SyntaxError:
        return None
    lambda_node = next((node for node in ast.walk(source_ast) if isinstance(node, ast.Lambda)), None)
    if lambda_node is None:  # could be a single line `def fn(x): ...`
        return None

    # HACK: Since we can (and most likely will) get source lines
    # where lambdas are just a part of bigger expressions, they will have
    # some trailing junk after their definition.
    #
    # Unfortunately, AST nodes only keep their _starting_ offsets
    # from the original source, so we have to determine the end ourselves.
    # We do that by gradually shaving extra junk from after the definition.
    lambda_text = source_text[lambda_node.col_offset :]
    lambda_body_text = source_text[lambda_node.body.col_offset :]
    min_length = len("lambda:_")  # shortest possible lambda expression
    while len(lambda_text) > min_length:
        try:
            # What's annoying is that sometimes the junk even parses,
            # but results in a *different* lambda. You'd probably have to
            # be deliberately malicious to exploit it but here's one way:
            #
            #     bloop = lambda x: False, lambda x: True
            #     get_short_lamnda_source(bloop[0])
            #
            # Ideally, we'd just keep shaving until we get the same code,
            # but that most likely won't happen because we can't replicate
            # the exact closure environment.
            code = compile(lambda_body_text, "<unused filename>", "eval")

            # Thus the next best thing is to assume some divergence due
            # to e.g. LOAD_GLOBAL in original code being LOAD_FAST in
            # the one compiled above, or vice versa.
            # But the resulting code should at least be the same *length*
            # if otherwise the same operations are performed in it.
            if len(code.co_code) == len(lambda_func.__code__.co_code):
                return lambda_text
        except SyntaxError:
            pass
        lambda_text = lambda_text[:-1]
        lambda_body_text = lambda_body_text[:-1]

    return None


lambda_pattern = r"\s*(\w+)\s*=\s*(lambda.*)"
lambda_regex = re.compile(lambda_pattern)

# TODO parsing functions with multiline decoration
def get_callable_source_code(f: Callable) -> str:
    """
    Return the source code, even if it's lambda function.
    """
    # Check for built in functions as their source code can not be fetched
    if isbuiltin(f) or (isclass(f) and getattr(f, "__module__", "") == "builtins"):
        raise InvalidArgument(
            f"Error getting source code for transform. {f} is a built-in function.\n"
            f"Please wrap in lambda like so: `lambda x: {f.__name__}(x)`"
        )

    try:
        source_lines, _ = getsourcelines(f)
    except (OSError, TypeError) as exc:
        # functions defined in pdb / REPL / eval / some other way in which
        # source code not clear
        error = (
            f"Cannot parse function {f.__name__}. This can occur if python cannot find the file the function was originally declared in.\n"
            "This can occur commonly the function's file is mounted to the filesystem (through a docker volume).\n"
            "To solve this copy the file into the container."
        )
        raise ValueError(error) from exc

    if isinstance(f, LambdaType) and f.__name__ == "<lambda>":
        source_code = get_short_lambda_source(f)
        if source_code:
            return source_code

    # If we have regular def function() type function we need to parse out any decorators
    # to do this we parse the function with ast and remove any lines that are not relevant
    source_code = dedent(''.join(source_lines))
    source_ast = ast.parse(source_code)
    # Get the first FunctionDef node by walking through the parsed ast tree
    function_node = next((node for node in ast.walk(source_ast) if isinstance(node, ast.FunctionDef)), None)
    if function_node is None:
        raise ValueError(f"Can't parse function:\n{source_code}")
    src = ''.join(source_lines[function_node.lineno - 1 : function_node.end_lineno])
    src = dedent(src).strip()
    if len(source_lines) > 1 and src[:3] == "def":
        return src

    match = lambda_regex.match(src)
    if match:
        func_name, func = match.groups()
        func_name = func_name.strip()
        func = func.strip()
    else:
        raise ValueError(f"Can't parse lambda:\n{src}")

    # Slice off trailing chars until we get a callable function
    try:
        eval_func = eval(func)
    except SyntaxError:
        raise ValueError(f"Cannot parse the source code due to syntax error:\n#####\n{func}")
    assert callable(eval_func)
    return func


def func_from_callable(func_: Callable[..., Output], env: Optional[Environment] = None) -> Func[Output]:
    """Generate a func from a variety of possible input data types.

    Args:
        func_ (Callable[..., Output]): Python function to convert to Func object
        env (Optional[Env], optional): Env object to include on the Function. Defaults to None.

    Returns:
        Func[Output]: Output Func object wrapped around the input function
    """
    if isinstance(env, dict):
        env = Environment.from_dict(env)
    elif env is None or env is Undefined:
        env = Environment()
    if isinstance(func_, Func):
        # assert not getattr(env,'imports',False)
        func: Func[Output] = Func(src=func_.src, function=function, env=env)
        return func
    else:
        assert callable(func_), f"tried to instantiate Func, but not callable {type(func_)}"

        func = Func(src=get_callable_source_code(func_), function=func_, env=env)
        return func



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import subprocess
from enum import Enum
from pathlib import Path
from typing import Optional

import typer

import dbgen.cli.styles as styles
from dbgen import __version__
from dbgen.cli.etl_step import etl_step_app
from dbgen.cli.model import model_app
from dbgen.cli.new import new_app
from dbgen.cli.options import chdir_option, config_option
from dbgen.cli.run import run_app
from dbgen.configuration import config, get_connections
from dbgen.utils.misc import which

app = typer.Typer(no_args_is_help=True)
app.add_typer(
    etl_step_app, name='etl-step', help="Commands related to running and monitoring single ETLSteps."
)
app.add_typer(run_app, name='run', help="Run DBgen models and monitor their status.")
app.add_typer(model_app, name='model', help="Validate, serialize and export DBgen models.")
app.add_typer(new_app, name='new', help="Create new DBgen models from templates.")

app.command("version", help="Print the version of dbgen")(lambda: styles.console.print(styles.LOGO_STYLE))


@app.command()
def version(short: bool = typer.Option(False, '-s', '--short', help='Only output the semantic version.')):
    """Print the dbgen version."""
    if short:
        styles.console.print(__version__)
    else:
        styles.console.print(styles.LOGO_STYLE)


@app.command(name="config")
def get_config(
    config_file: Path = config_option,
    show_password: bool = False,
    show_defaults: bool = False,
    out_pth: Optional[Path] = typer.Option(
        None,
        '--out',
        '-o',
        help="Location to write parametrized config",
    ),
    _chdir: Path = chdir_option,
):
    """
    Prints out the configuration of dbgen given an optional config_file or using the envvar DBGEN_CONFIG
    """
    styles.theme_typer_print(styles.LOGO_STYLE)
    # If out_pth provided write the current config to the path provided and return
    if out_pth:
        with open(out_pth, "w") as f:
            f.write(config.display(True, True))

    typer.echo(config.display(show_defaults, show_password))


class DBgenDatabase(str, Enum):
    """enum for setting database to connect to"""

    META = "meta"
    MAIN = "main"


@app.command(name="connect")
def test_conn(
    connect: DBgenDatabase = typer.Argument(DBgenDatabase.MAIN, help="Database to connect the meta or main."),
    config_file: Optional[Path] = config_option,
    test: bool = typer.Option(False, "-t", "--test", help="Test the main and metadb connections"),
    show_password: bool = typer.Option(
        False, "-p", "--show-password", help="Expose password in printed dsn when testing."
    ),
):
    """
    Prints out the configuration of dbgen given an optional config_file or using the envvar DBGEN_CONFIG
    """
    main_conn, meta_conn = get_connections()

    # If connect is chosen connect to the database selected using CLI sql
    if connect and not test:
        # set the engine based on connect string provided
        conn = main_conn if connect == DBgenDatabase.MAIN else meta_conn
        # Test connection to database
        if not conn.test():
            styles.bad_typer_print(f"Cannot connect to {str(connect)} db")
            raise typer.Exit(2)
        # Attempt to use psql and pgcli to connect to database
        try:
            # Filter out executibles using which function
            exes = list(filter(lambda x: x is not None, map(which, ("pgcli", "psql"))))
            # If we find no executables exit
            if not exes:
                styles.bad_typer_print(
                    "Cannot find either psql or pgcli in $PATH. Please install them to connect to database."
                )
                raise typer.Exit(2)
            # If we have valid executible run the command with the dsn provided
            command = exes[0]
            if command is not None and not config.testing:
                subprocess.check_call(
                    [command, conn.url(False, True)],
                )
        except subprocess.CalledProcessError as exc:
            styles.bad_typer_print("Error connecting!")
            styles.bad_typer_print(str(exc))
        # Quit once finished
        raise typer.Exit()

    styles.delimiter()
    failed = False
    for conn, label in zip((main_conn, meta_conn), ("Main", "Meta")):
        styles.good_typer_print(f"Checking {label} DB...")
        check = conn.test()
        if check:
            styles.good_typer_print(
                f"Connection to {label} DB at {conn.url(not show_password,True)} all good!"
            )
        else:
            styles.bad_typer_print(f"Cannot connect to {label} DB at {conn.url(not show_password,True)}!")
            failed = True
        styles.delimiter()
    raise typer.Exit(code=2 if failed else 0)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import traceback
from bdb import BdbQuit
from typing import Any, Dict, Generator, List, Optional, Sequence, Tuple
from uuid import UUID

from pydantic import PrivateAttr
from pydasher import hasher

from dbgen._enum import RunStatus
from dbgen.core.base import Base, encoders
from dbgen.core.etl_step import ETLStep
from dbgen.core.model_settings import BaseModelSettings
from dbgen.core.node.extract import Extract
from dbgen.core.run.utilities import RunConfig


class TestRunResults(Base):
    number_of_extracted_rows: int = 0
    status: RunStatus = RunStatus.initialized


class ETLStepTestRunner(Base):
    _etl_step: ETLStep = PrivateAttr(None)

    def test(
        self,
        etl_step: ETLStep,
        test_rows: Optional[Sequence[Any]] = None,
        settings: Optional[BaseModelSettings] = None,
    ):
        self._etl_step = etl_step
        run_config: RunConfig = RunConfig(settings=settings) if settings else RunConfig()
        extract = etl_step.extract
        results = TestRunResults()
        extract._set_run_config(run_config)
        with extract:
            # Start while loop to iterate through the nodes
            self._logger.debug('Looping through extracted rows...')

            for batch_ind, batch in enumerate(self.batchify(extract, 1)):
                (
                    _,
                    rows_to_load,
                    rows_processed,
                    inputs_skipped,
                    output,
                    exc,
                ) = self.transform_batch(batch, run_config)
                # Check if transforms or loads raised an error
                self._logger.info(f'Batch {batch_ind} processed')
                self._logger.info(f'Skipped {inputs_skipped} rows int batch')
                self._logger.info(f'Processed {rows_processed} rows int batch')
                self._logger.info(f'Namespace: {output}')
                if exc:
                    msg = f"Error when running etl_step {self._etl_step.name}"
                    self._logger.error(msg)
                    results.status = RunStatus.failed
                    return results
                self._fake_load(etl_step, rows_to_load)
                results.number_of_extracted_rows += len(batch)
                results.status = RunStatus.completed
        return results

    def _fake_load(self, etl_step: ETLStep, rows_to_load: Dict[str, Dict[UUID, Any]]):
        for load in etl_step._sorted_loads():
            self._logger.debug(f'Loading into {load}')
            rows = rows_to_load[load.hash]
            for row in rows:
                self._logger.info(f'Attempting to load row: {row}')

    def transform_batch(self, batch: List[Tuple[UUID, Dict[str, Dict[str, Any]]]], run_config: 'RunConfig'):
        """Transform a batch of extracted namespaces."""
        # initialize the master dict for the rows that will need to be loaded after
        rows_to_load: Dict[str, Dict[UUID, dict]] = {node.hash: {} for node in self._etl_step.loads}
        processed_hashes = []
        inputs_skipped = 0
        for input_hash, row in batch:
            try:
                output, skipped = self._etl_step._transform(row, rows_to_load, run_config)
                if not skipped:
                    processed_hashes.append(input_hash)
                else:
                    inputs_skipped += 1
            except (KeyboardInterrupt, SystemExit, BdbQuit):
                raise
            except BaseException:
                return None, None, None, inputs_skipped, None, traceback.format_exc()
        return processed_hashes, rows_to_load, len(batch), inputs_skipped, output, None

    def batchify(
        self,
        extract: Extract,
        batch_size: int,
    ) -> Generator[Any, None, None]:
        # initialize the batch and counts
        batch = []
        # Loop the the rows in the extract function
        for row in extract.extract():
            # Check the hash of the inputs against the metadatabase
            processed_row = extract.process_row(row)
            input_hash = self._get_hash(processed_row, self._etl_step.uuid)
            batch.append((input_hash, {extract.hash: processed_row}))

            # If batch size is reached advance bar and yield
            if len(batch) == batch_size:
                yield batch
                batch = []

    def _get_hash(self, extracted_dict: Dict[str, Any], etl_step_uuid: UUID) -> UUID:
        # Convert Row to a dictionary so we can hash it for repeat-checking
        input_hash = UUID(hasher((etl_step_uuid, extracted_dict), encoders=encoders))
        return input_hash



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import TYPE_CHECKING, Any, Callable, Dict
from typing import Generator
from typing import Generator as GenType
from typing import Mapping, Optional, TypeVar, Union

from pydantic import Field, PrivateAttr, root_validator, validator

from dbgen.core.func import Environment, Func, func_from_callable
from dbgen.core.node.computational_node import ComputationalNode

if TYPE_CHECKING:
    from dbgen.core.run.utilities import RunConfig
extractor_type = GenType[Dict[str, Mapping[str, Any]], None, None]

Output = TypeVar('Output')
T = TypeVar('T')
T1 = TypeVar('T1')

# TODO remove **_ from setup and fix tests
class Extract(ComputationalNode[T]):
    """
    Base Class for all extraction steps.

    Must implement the extract method which returns a list of rows to map the transforms over.

    Must also be subscriptable to connect to the transforms when coding a model.
    """

    _extractor: GenType[Union[Dict[str, Any], T], None, None] = PrivateAttr(None)
    _run_config: 'RunConfig' = PrivateAttr(None)

    class Config:
        """Pydantic Config"""

        underscore_attrs_are_private = True

    # Overwrite these when writing custom extractor
    def setup(self) -> None:
        pass

    def extract(self: 'Extract[T]') -> GenType[T, None, None]:
        yield {}  # type: ignore

    def teardown(self) -> None:
        pass

    def length(self) -> Optional[int]:
        return None

    # Internal Do not Overwrite

    def _set_run_config(self, run_config: 'RunConfig'):
        self._run_config = run_config

    def run(
        self, _: Dict[str, Mapping[str, Any]], run_config: Optional['RunConfig'] = None
    ) -> Optional[Dict[str, Any]]:
        row = next(self._extractor)
        return self.process_row(row)

    def process_row(self, row):
        if row is None:
            return {}
        elif isinstance(row, Mapping):
            return dict(row)
        elif isinstance(row, tuple):
            l1, l2 = len(row), len(self.outputs)
            if l1 != l2:
                raise ValueError(
                    f"Expected {l2} output from extract {self}, but got {l1} outputs.\n"
                    f"If you intended to return a length {l1} tuple as the single output, please wrap the tuple in a singleton tuple\n"
                    "like so 'yield (tuple_to_return,)'"
                )
            return {name: val for name, val in zip(self.outputs, row)}
        else:
            if len(self.outputs) != 1:
                raise ValueError(
                    f"{self} expected multiple outputs but output a {type(row)} which cannot have its length measured"
                )
            return {list(self.outputs)[0]: row}

    def __enter__(self):
        """Call setup when extract used in with block."""
        self.setup()

    def __exit__(self, *_):
        """Call teardown when exiting extract with block regardless of error."""
        self.teardown()

    def __str__(self):
        return f"{self.__class__.__qualname__}<outputs= {self.outputs}>"


# TODO add better error messaging when user passes in a non-arg to pyblock
class PythonExtract(Extract[Output]):

    env: Optional[Environment] = Field(default_factory=lambda: Environment(imports=set()))
    function: Func[Output]

    @validator('function', pre=True)
    def convert_callable_to_func(cls, function: Union[Func[Output], Callable[..., Output]], values):
        env = values.get('env', Environment())
        if isinstance(function, (Func, dict)):
            return function
        elif callable(function):
            return func_from_callable(function, env=env)
        raise ValueError(f"Unknown function type {type(function)} {function}")

    @root_validator
    def check_nargs(cls, values):
        if "function" not in values:
            return values
        inputs = values.get("inputs")
        func = values.get("function")
        num_req_args = values.get("function").number_of_required_inputs
        num_max_args = values.get("function").number_of_inputs
        if inputs:
            number_of_inputs = len(inputs)
            assert (
                number_of_inputs >= num_req_args
            ), f"Too few arguments supplied to Func. Number of Inputs: {number_of_inputs}, Number of Args: {num_req_args}\n{values}"
            if not func.var_positional_keyword:
                assert (
                    number_of_inputs <= num_max_args
                ), f"Too many arguments supplied to Func. Number of Inputs: {number_of_inputs}, Max Number of Args: {num_max_args}\n"
        return values

    def extract(self) -> Generator[Output, None, None]:
        inputvars = self._get_inputs({})
        args = {key: val for key, val in inputvars.items() if key.isdigit()}
        kwargs = {key: val for key, val in inputvars.items() if key not in args}
        if 'settings' in self.function.argnames:
            setting_index = self.function.argnames.index('settings')
            if 'settings' not in kwargs or len(args) < setting_index + 1:
                kwargs['settings'] = self._run_config.settings

        yield from self.function(*args.values(), **kwargs)  # type: ignore



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import TYPE_CHECKING, Any, Dict
from typing import Generator as GenType
from typing import Optional, Type, TypeVar, Union, overload

from pydantic import Field
from sqlalchemy import text
from sqlalchemy.dialects import postgresql
from sqlmodel.sql.expression import Select, SelectOfScalar

from dbgen.core.dependency import Dependency
from dbgen.core.node.extract import Extract
from dbgen.core.statement_parsing import _get_select_keys, get_statement_dependency
from dbgen.utils.sql import Connection

if TYPE_CHECKING:
    from psycopg import AsyncConnection  # pragma: no cover
    from sqlalchemy.engine import Connection as SAConnection  # pragma: no cover


SCHEMA_DEFAULT = "public"

T = TypeVar('T')

postgresql_dialect = postgresql.dialect()  # type: ignore


class BaseQuery(Extract[T]):
    query: str
    params: Dict[str, Any] = Field(default_factory=dict)
    dependency: Dependency = Field(default_factory=Dependency)
    _connection: 'SAConnection'
    _yield_per: Optional[int] = None

    def _get_dependency(self) -> Dependency:
        return self.dependency

    @classmethod
    def from_select_statement(
        cls, select_statement: Select[T], connection: Connection = None, **kwargs
    ) -> "BaseQuery[T]":
        columns, tables, fks = get_statement_dependency(select_statement)
        outputs = _get_select_keys(select_statement)
        get_table_name = lambda x: f"{x.schema}.{x.name}" if getattr(x, "schema", "") else x.name
        dependency = Dependency(
            tables_needed=[get_table_name(x) for x in tables],
            columns_needed=[f"{get_table_name(x.table)}.{x.name}" for x in columns.union(fks)],
        )
        compiled_statement = select_statement.compile()
        return cls(
            inputs=[],
            outputs=outputs,
            query=str(compiled_statement),
            params=compiled_statement.params,
            dependency=dependency,
        )

    def render_query(self) -> str:
        """Stringifies the query with the bound parameters."""
        compiled_query = (
            text(self.query).bindparams(**self.params).compile(compile_kwargs={'literal_binds': True})
        )
        return str(compiled_query)

    def set_connection(self, connection: 'SAConnection', yield_per: Optional[int] = None):
        self._connection = connection
        self._yield_per = yield_per

    def length(self) -> int:
        rows: int = self._connection.execute(text(self.count_statement)).scalar()  # type: ignore
        return rows

    async def _async_length(self, *, connection: 'AsyncConnection' = None, **_) -> Optional[int]:
        assert connection
        result = await connection.execute(
            self.count_statement,
            self.params,
        )
        out = await result.fetchone()
        (count,) = out if out else (None,)
        return count

    @property
    def compiled_query(self):
        return str(text(self.query).compile(dialect=postgresql_dialect))

    @property
    def count_statement(self):
        return str(
            text(f'select count(1) from ({self.render_query()}) as X').compile(dialect=postgresql_dialect)
        )

    def extract(
        self,
    ) -> GenType[T, None, None]:
        if self._yield_per:
            result = self._connection.execution_options(stream_results=True).execute(
                text(self.query).bindparams(**self.params)
            )
            while chunk := result.fetchmany(self._yield_per):
                for row in chunk:
                    yield dict(row)  # type: ignore
        else:
            result = self._connection.execute(text(self.query).bindparams(**self.params))
            yield from result.mappings()  # type: ignore


class ExternalQuery(BaseQuery[T]):
    connection: Connection

    @classmethod
    def from_select_statement(
        cls, select_statement: Select, connection: 'Connection' = None, **kwargs
    ) -> "ExternalQuery":
        selected_keys = _get_select_keys(select_statement)
        return cls(
            query=str(select_statement),
            outputs=selected_keys,
            connection=connection,
        )

    def setup(self):
        engine = self.connection.get_engine()
        self._connection = engine.connect()

    def teardown(self):
        self._connection.close()


@overload
def Query(select_statement: SelectOfScalar[T], connection: Connection) -> ExternalQuery[T]:
    ...


@overload
def Query(select_statement: SelectOfScalar[T], connection: None = None) -> BaseQuery[T]:
    ...


@overload
def Query(select_statement: Select[T], connection: Connection) -> ExternalQuery[T]:
    ...


@overload
def Query(select_statement: Select[T], connection: None = None) -> BaseQuery[T]:
    ...


def Query(select_statement, connection: Optional[Connection] = None) -> Union[BaseQuery, ExternalQuery]:
    cls: Union[Type[BaseQuery], Type[ExternalQuery]] = BaseQuery if connection is None else ExternalQuery
    return cls.from_select_statement(select_statement, connection=connection)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Optional, Union, cast

import sqlalchemy as sa
from pydantic import Field
from pydantic.networks import PostgresDsn
from pydantic.tools import parse_obj_as
from pydantic.types import SecretStr
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext import compiler
from sqlalchemy.schema import DDLElement, PrimaryKeyConstraint

from dbgen.core.base import Base
from dbgen.utils.type_coercion import json_dumps

SCHEMA_DEFAULT = 'public'


class Connection(Base):
    scheme: str = "postgresql"
    user: str = "postgres"
    password: Optional[SecretStr] = None
    host: str = "localhost"
    port: int = 5432
    database: str = "dbgen"
    schema_: str = Field(SCHEMA_DEFAULT, alias="schema")
    _hashexclude_ = {"password"}

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.url()

    @classmethod
    def from_uri(
        cls, uri: Union[PostgresDsn, str], schema: str = SCHEMA_DEFAULT, password: Optional[SecretStr] = None
    ) -> "Connection":
        if isinstance(uri, str):
            uri = parse_obj_as(PostgresDsn, uri)
            uri = cast(PostgresDsn, uri)

        assert uri.path, f"uri is missing database {uri}"
        return cls(
            host=uri.host or "localhost",
            user=uri.user,
            password=password.get_secret_value() if password else uri.password,
            port=uri.port or 5432,
            database=uri.path.lstrip("/"),
            schema=schema,
        )

    def url(self, mask_password: bool = True, with_schema: bool = False):
        if mask_password:
            password = "******"
        elif self.password:
            password = self.password.get_secret_value()
        else:
            password = ""
        schema_suffix = f"?options=--search_path%3d{self.schema_}" if with_schema else ""
        return (
            f"{self.scheme}://{self.user}:{password}@{self.host}:{self.port}/{self.database}{schema_suffix}"
        )

    def get_engine(self, echo: bool = False):
        return create_engine(
            url=self.url(mask_password=False),
            connect_args={"options": f"-csearch_path={self.schema_}"},
            json_serializer=json_dumps,
            echo=echo,
        )

    def test(self):
        engine = self.get_engine()
        try:
            with engine.connect() as conn:
                conn.execute("select 1")
        except OperationalError as exc:
            self._logger.error(f"Trouble connecting to database at {self}.\n{exc}")
            return False
        return True


class CreateMaterializedView(DDLElement):
    def __init__(self, name, selectable):
        self.name = name
        self.selectable = selectable


@compiler.compiles(CreateMaterializedView)
def compile(element, compiler, **kw):
    # Could use "CREATE OR REPLACE MATERIALIZED VIEW..."
    # but I'd rather have noisy errors
    return "CREATE MATERIALIZED VIEW {} AS {}".format(
        element.name,
        compiler.sql_compiler.process(element.selectable, literal_binds=True),
    )


class CreateView(DDLElement):
    def __init__(self, name, selectable, schema):
        self.name = name
        self.selectable = selectable
        self.selectable = selectable
        self.schema = schema


@compiler.compiles(CreateView)
def compile_view(element, compiler, **kw):
    return "CREATE OR REPLACE VIEW {}.{} AS {}".format(
        element.schema,
        element.name,
        compiler.sql_compiler.process(element.selectable, literal_binds=True),
    )


def create_mat_view(name, selectable, metadata: sa.MetaData):
    _mt = sa.MetaData()  # temp metadata just for initial Table object creation
    t = sa.Table(name, _mt)  # the actual mat view class is bound tosa.metadata
    for c in selectable.c:
        t.append_column(sa.Column(c.name, c.type, primary_key=c.primary_key))

    if not (any([c.primary_key for c in selectable.c])):
        t.append_constraint(PrimaryKeyConstraint(*[c.name for c in selectable.c]))

    sa.event.listen(metadata, "after_create", CreateMaterializedView(name, selectable))

    @sa.event.listens_for(metadata, "after_create")
    def create_indexes(target, connection, **kw):
        for idx in t.indexes:
            idx.create(connection)

    sa.event.listen(metadata, "before_drop", sa.DDL("DROP MATERIALIZED VIEW IF EXISTS " + name))
    return t


def create_view(
    name,
    selectable,
    schema,
    metadata: sa.MetaData,
    columns: Optional[List[sa.Column]] = None,
):
    _mt = sa.MetaData(schema=schema)  # temp metadata just for initial Table object creation
    t = sa.Table(name, _mt)  # the actual mat view class is bound tosa.metadata
    if columns is None and getattr(selectable, "c", None):
        columns = [sa.Column(c.name, c.type, primary_key=c.primary_key) for c in selectable.c]

    assert columns
    for c in columns:
        t.append_column(c)

    if not (any([c.primary_key for c in columns])):
        t.append_constraint(PrimaryKeyConstraint(*[c.name for c in columns]))

    sa.event.listen(metadata, "after_create", CreateView(name, selectable, schema))

    @sa.event.listens_for(metadata, "after_create")
    def create_indexes(target, connection, **kw):
        for idx in t.indexes:
            idx.create(connection)

    sa.event.listen(metadata, "before_drop", sa.DDL(f"DROP VIEW IF EXISTS {schema}.{name}"))
    return t


def refresh_mat_view(name, concurrently):
    # since session.execute() bypasses autoflush, must manually flush in order
    # to include newly-created/modified objects in the refresh
    sa.session.flush()
    _con = "CONCURRENTLY " if concurrently else ""
    sa.session.execute("REFRESH MATERIALIZED VIEW " + _con + name)


def refresh_all_mat_views(concurrently=True):
    """Refreshes all materialized views. Currently, views are refreshed in
    non-deterministic order, so view definitions can't depend on each other."""
    mat_views = sa.inspect(sa.engine).get_view_names(include="materialized")
    for v in mat_views:
        refresh_mat_view(v, concurrently)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import contextlib
import logging
from enum import Enum
from io import StringIO
from logging import Formatter, Logger
from logging.handlers import RotatingFileHandler
from pathlib import Path
from textwrap import dedent
from typing import Tuple

# Add error message as this is first module imported by dbgen
try:
    from rich.console import Console
    from rich.logging import RichHandler
    from rich.traceback import install
except ModuleNotFoundError:
    print(
        dedent(
            """
    ERROR: It appears that rich, a required dbgen python dependency, is not installed. This is commonly caused by DBgen not being installed correctly. \n\
    Please reference https://dbgen.modelyst.com/pages/common_errors/#missing-module-error for more details. The full error can be seen below.
    """
        )
    )
    raise
install()

logging_console = Console()


class LogLevel(str, Enum):
    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    CRITICAL = 'CRITICAL'

    def get_log_level(self):
        return getattr(logging, self)


def capture_stdout(func):
    def wrapped(*args, **kwargs):
        logger = logging.getLogger(f"dbgen.pyblock.{func.name}")
        stream = StringIO()
        with contextlib.redirect_stdout(stream):
            output = func(*args, **kwargs)
        stdout = stream.getvalue().strip()
        if stdout:
            logger.debug(stdout)
        return output

    return wrapped


def setup_logger(
    level: LogLevel = LogLevel.DEBUG, std_out_level: LogLevel = LogLevel.INFO
) -> Tuple[Logger, RichHandler]:
    std_out_level = std_out_level or level
    custom_logger = logging.getLogger("dbgen")
    custom_logger.propagate = True
    custom_logger.setLevel(level.get_log_level())
    rich_handler = add_stdout_logger(custom_logger, std_out_level)
    return custom_logger, rich_handler


def add_stdout_logger(logger, stdout_level: LogLevel = LogLevel.DEBUG) -> RichHandler:
    rich_handler = RichHandler(level=stdout_level.get_log_level(), markup=True, console=logging_console)
    log_format = r"[magenta]\[%(name)s][/magenta] - %(message)s"
    rich_handler.setFormatter(Formatter(log_format))
    logger.addHandler(rich_handler)
    return rich_handler


def add_file_handler(logger: Logger, level: LogLevel = LogLevel.DEBUG, file_name: Path = None):
    file_handler = RotatingFileHandler(str(file_name))
    log_format = "[%(asctime)s] - %(name)s - %(levelname)s - %(message)s"
    file_handler.setFormatter(Formatter(log_format))
    file_handler.setLevel(level.get_log_level())
    logger.addHandler(file_handler)
    return file_handler



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime
from logging import getLogger
from pathlib import Path
from typing import List, Optional
from uuid import UUID

import typer
from rich.console import Console
from rich.table import Column, Table
from rich.text import Text
from sqlmodel import Session, select

import dbgen.cli.styles as styles
import dbgen.exceptions as exceptions
from dbgen.cli.options import (
    chdir_option,
    config_option,
    log_file_option,
    model_string_option,
    version_option,
)
from dbgen.cli.queries import get_runs
from dbgen.cli.utils import confirm_build, set_confirm, test_connection, validate_model_str
from dbgen.configuration import config, get_connections, root_logger, stdout_handler
from dbgen.core.metadata import ETLStepRunEntity, ModelEntity, RunEntity, Status
from dbgen.core.run.utilities import RunConfig
from dbgen.utils.log import LogLevel, add_file_handler

run_app = typer.Typer(name='run')
logger = getLogger(__name__)


@run_app.command('status')
def status(
    run_id: Optional[int] = None,
    last: int = 1,
    all_runs: bool = typer.Option(False, '-a', '--all', help='show all runs ever'),
    error: bool = typer.Option(False, '--error', help='show error for failed runs'),
    query: bool = typer.Option(False, '--query', help='show error for failed runs'),
    statuses: List[str] = typer.Option(
        [], '-s', '--status', help='Filter out only statuses that match input val'
    ),
    _config_file: Path = config_option,
):
    """Print a table summarizing DBgen Model runs."""
    column_dict = {'error': Column('Error'), 'status': Column('Status', style='green')}
    _, meta_conn = get_connections()
    test_connection(meta_conn)
    meta_engine = meta_conn.get_engine()
    filtered_keys = {'etl_step_id', 'created_at'}
    if not error:
        filtered_keys.add('error')
    if not query:
        filtered_keys.add('query')
    columns = [
        'name',
        *(c.name for c in ETLStepRunEntity.__table__.c if c.name not in filtered_keys),
    ]
    show_cols = {col: col.replace('number_of_', '') for col in columns}
    table = Table(
        *[column_dict.get(col, col) for col in show_cols],
        title='run',
        show_lines=True,
        highlight=True,
        border_style='magenta',
    )
    status_map = {
        'failed': 'white on red',
        'upstream_failed': 'black on yellow',
        'completed': 'green',
        'excluded': 'blue',
    }

    def style(key, col):
        if key == 'status':
            return Text(str(col), style=status_map.get(col))
        elif key == 'memory_usage':
            if col and col > 1024:
                return f"{(col/1024):3.1f} GB"
            return f"{col:3.1f} MB" if col else ''
        elif key == 'runtime':
            if not col:
                return 'N/A'
            hours = int(col / 3600)
            minutes = int((col % 3600) / 60)
            seconds = col % 3600 % 60
            output = ''
            output += f"[blue]{hours:01.0f}[/blue]:"
            output += f"[magenta]{minutes:02.0f}[/magenta]:"
            output += f"[white]{seconds:1.3f}[/white]"

            return output
        return str(col)

    for row in get_runs(run_id, meta_engine, all_runs, statuses, last):
        table.add_row(
            *[style(key, row[key]) for (key, col) in show_cols.items()],
        )
    console = Console()
    console.print(table)


@run_app.callback(invoke_without_command=True)
def run_model(
    ctx: typer.Context,
    model_str: str = model_string_option,
    include: List[str] = typer.Option([], help="ETLSteps to include"),
    exclude: List[str] = typer.Option([], help="ETLSteps to xclude"),
    retry: bool = typer.Option(False, help="Ignore repeat checking"),
    start: Optional[str] = typer.Option(None, help="ETLStep to start run at"),
    until: Optional[str] = typer.Option(None, help="ETLStep to finish run at."),
    build: bool = typer.Option(
        None,
        help="Delete the entire db and meta schema and rebuild from scratch.",
        callback=confirm_build,
    ),
    level: LogLevel = typer.Option(LogLevel.INFO, help='Use the RemoteETLStep Runner'),
    bar: bool = typer.Option(True, help="Show progress bar"),
    skip_row_count: bool = typer.Option(False, help="Show progress bar"),
    skip_on_error: bool = typer.Option(False, help="Skip a row in etl_step on error"),
    fast_fail: bool = typer.Option(
        False, '--fast-fail', help="Exclude all downstream ETLSteps once one has failed."
    ),
    fail_downstream: bool = typer.Option(
        False, '--fail-downstream', help="Exclude all dependent ETLSteps once an ETLStep fails"
    ),
    batch: Optional[int] = typer.Option(None, help="Batch size for all etl_steps in run."),
    batch_number: int = typer.Option(10, help="Default number of batches per etl_step."),
    pdb: bool = typer.Option(False, '--pdb', help="Drop into pdb on breakpoints"),
    log_file: Path = log_file_option,
    log_file_level: LogLevel = typer.Option(
        None, help="Log Level to write to the --log-file if set. (Defaults to --level if not set)"
    ),
    remote: bool = typer.Option(False, help='Use the RemoteETLStep Runner'),
    run_async: bool = typer.Option(False, '--async', help='Use the RemoteGenerator Runner'),
    config_file: Path = config_option,
    no_conf: bool = typer.Option(
        False,
        "--no-confirm",
        "-y",
        is_eager=True,
        callback=set_confirm,
    ),
    rerun_failed: bool = typer.Option(
        False,
        help="Only rerun the etl_steps that failed or were excluded in the previous run.",
    ),
    version: bool = version_option,
    _chdir: Path = chdir_option,
):
    """Run a model."""
    if ctx.invoked_subcommand is not None:
        return
    # Start connection from config
    main_conn, meta_conn = get_connections()
    # Use config model_str if none is provided
    if model_str is None:
        model_str = config.model_str
    # validate the model_str
    model = validate_model_str(model_str)
    styles.delimiter(styles.typer.colors.GREEN)
    styles.good_typer_print(f"Running model [theme]{model.name!r}[/theme]...")
    styles.delimiter(styles.typer.colors.GREEN)
    # Initialize the run Configuration
    run_config = RunConfig(
        retry=retry,
        start=start,
        until=until,
        exclude=exclude,
        include=include,
        progress_bar=bar,
        log_level=level,
        skip_row_count=skip_row_count,
        fast_fail=fast_fail,
        fail_downstream=fail_downstream,
        skip_on_error=skip_on_error,
        batch_size=batch,
        batch_number=batch_number,
    )
    # Set the stdout logger to the --level value
    stdout_handler.setLevel(run_config.log_level.get_log_level())
    # If --log-file is add a file handler to the logger
    if log_file:
        # Default to --level if --log-file-level is not set
        log_file_level = log_file_level or run_config.log_level
        add_file_handler(root_logger, log_file_level, log_file)
    # Validate the run configuration parameters
    invalid_run_config_vals = run_config.get_invalid_markers(model)
    if invalid_run_config_vals:
        lines = [f"{name}: {' ,'.join(val)}" for name, val in invalid_run_config_vals.items()]
        error = "\n".join(lines)
        raise typer.BadParameter(f"Invalid run configuration parameters passed:\n{error}")

    # Test each connection with simple select command
    logger.debug('testing connections')
    for name, conn in (('main', main_conn), ('meta', meta_conn)):
        logger.debug(f'testing {name} connection...')
        test_connection(conn, name)
    # Grab engine from each connection
    main_engine, meta_engine = main_conn.get_engine(), meta_conn.get_engine()
    # If we turn on pdb mode we need to turn off the progress bar
    if pdb:
        config.pdb = pdb
        run_config.progress_bar = not pdb
    # Pass all the arguments to the model run command
    try:
        out_run = model.run(
            main_engine,
            meta_engine,
            run_config,
            build=build,
            rerun_failed=rerun_failed,
            remote=remote,
            run_async=run_async,
        )
    except exceptions.SerializationError as exc:
        raise typer.BadParameter(
            f"A etl_step in your model could not be deserialized. Did you use a custom extractor?"
        ) from exc
    except exceptions.DBgenException as exc:
        logger.exception(exc)
        raise typer.Exit(code=1)

    # Once run is done talk to the meta database to report to the user how the run went
    with Session(meta_engine) as session:
        run = session.exec(select(RunEntity).where(RunEntity.id == out_run.id)).one()
        failed_etl_steps = [
            etl_step_run.etl_step.name
            for etl_step_run in run.etl_step_runs
            if etl_step_run.status == 'failed'
        ]
        excluded_etl_steps = [
            etl_step_run.etl_step.name
            for etl_step_run in run.etl_step_runs
            if etl_step_run.status in (Status.excluded, Status.upstream_failed)
        ]
        rows_inserted = sum(etl_step_run.rows_inserted for etl_step_run in run.etl_step_runs)
        rows_updated = sum(etl_step_run.rows_updated for etl_step_run in run.etl_step_runs)

    if run.errors:
        styles.delimiter(styles.typer.colors.RED)
        styles.bad_typer_print(f"Encountered {run.errors} during the run! The following etl_steps failed:")
        styles.bad_typer_print('\n'.join(map(repr, failed_etl_steps)))
    if run.status == Status.completed and run.runtime:
        styles.delimiter(styles.typer.colors.GREEN)
        styles.good_typer_print(
            f"Finished Running {len(model.etl_steps)} ETLStep(s) in {run.runtime.total_seconds():.3f}(s). {len(excluded_etl_steps)} Steps were Excluded."
        )
        styles.good_typer_print(
            f"The run attempted to insert {rows_inserted} and update {rows_updated} rows."
        )
    else:
        styles.delimiter(styles.typer.colors.RED)
        styles.bad_typer_print(
            f"Run failed! Only {len(run.etl_step_runs)} of {len(model.etl_steps)} ETLSteps run. {len(excluded_etl_steps)} Steps were Excluded."
        )
        raise typer.Exit(code=1)

    raise typer.Exit(code=0)


@run_app.command('initialize')
def run_initialize(model_id: UUID, config_file: Path = config_option):
    """Initialize a run."""
    # Notify of config file
    _, meta_conn = get_connections()
    test_connection(meta_conn)
    meta_engine = meta_conn.get_engine()
    typer.echo(config.display())
    with Session(meta_engine) as session:
        model = session.get(ModelEntity, model_id)
        if not model:
            raise ValueError(f"Invalid model id, could not fine model with ID {model_id}")
        model.last_run = datetime.now()
        run = RunEntity(status='initialized')
        session.add(run)
        session.commit()
        session.refresh(run)
    typer.echo(run.id)
    return



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Utilities for printing things to screen for CLI."""
from typing import Callable, Union

import typer
from rich.console import Console, Group, RenderableType
from rich.panel import Panel
from rich.text import Text
from rich.theme import Theme

from dbgen import __version__

THEME_COLOR = typer.colors.BRIGHT_MAGENTA
theme = Theme({'theme': 'magenta'})
console = Console(theme=theme)
LOGO = """
    ____  ____
   / __ \\/ __ )____ ____  ____
  / / / / __  / __ `/ _ \\/ __ \\
 / /_/ / /_/ / /_/ /  __/ / / /
/_____/_____/\\__, /\\___/_/ /_/
            /____/
"""

VERSION = f"""VERSION: {__version__}"""


def delimiter(color: str = THEME_COLOR):
    console.rule(style=color)


LOGO_STYLE = Panel.fit(
    Group(Panel(Text(LOGO)), Panel(Text(VERSION, justify='center'))),
    style='theme',
)


# Easy printers
typer_print: Callable[
    [str],
    Callable[
        [Union[str, RenderableType]],
        None,
    ],
] = lambda color: lambda msg: console.print(msg, style=color)
good_typer_print = typer_print('green')
bad_typer_print = typer_print('red')
theme_typer_print = typer_print('theme')
greens = lambda x: typer.style(x, fg=typer.colors.BRIGHT_GREEN)
reds = lambda x: typer.style(x, fg=typer.colors.BRIGHT_RED)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import Counter
from json import loads
from typing import TYPE_CHECKING, Dict, Generic, List, Tuple, TypeVar
from uuid import UUID

import sqlalchemy
from networkx import DiGraph
from pydantic import Field, PrivateAttr, validator
from sqlalchemy import MetaData, inspect, text
from sqlalchemy.future import Engine
from sqlalchemy.orm import registry as sa_registry

from dbgen.core.base import Base
from dbgen.core.context import ModelContext
from dbgen.core.entity import BaseEntity
from dbgen.core.etl_step import ETLStep
from dbgen.core.metadata import ModelEntity, RunEntity, meta_registry
from dbgen.core.model_settings import BaseModelSettings
from dbgen.exceptions import ModelRunError
from dbgen.utils.graphs import serialize_graph, topsort_with_dict

if TYPE_CHECKING:
    from dbgen.core.run.utilities import RunConfig  # pragma: no cover

SettingType = TypeVar('SettingType', bound=BaseModelSettings)


class Model(Base, Generic[SettingType]):
    name: str
    etl_steps: List[ETLStep] = Field(default_factory=list)
    registry: sa_registry = Field(default_factory=lambda: BaseEntity._sa_registry)
    meta_registry: sa_registry = Field(default_factory=lambda: meta_registry)
    settings: SettingType = Field(default_factory=lambda: BaseModelSettings())
    _context: ModelContext = PrivateAttr(None)
    _hashinclude_ = {"name", "etl_steps"}

    class Config:
        """Pydantic Config"""

        arbitrary_types_allowed = True

    def __enter__(self):
        self._context = ModelContext(context_dict={'model': self})
        return self._context.__enter__()['model']

    def __exit__(self, *args):
        self._context.__exit__(*args)
        del self._context
        self.validate(self)

    @validator("etl_steps")
    def unique_etl_step_names(cls, value: List[ETLStep]) -> List[ETLStep]:
        name_count = Counter([x.name for x in value])
        dup_names = {name for name, count in name_count.items() if count > 1}
        if dup_names:
            raise ValueError(f"Found duplicate etl_step names, all names must be unique: {dup_names}")
        return value

    @validator("etl_steps")
    def sort_etl_steps(cls, value: List[ETLStep]) -> List[ETLStep]:
        return value

    def etl_steps_dict(self) -> Dict[str, ETLStep]:
        return {etl_step.name: etl_step for etl_step in self.etl_steps}

    def add_etl_step(self, etl_step: ETLStep):
        if etl_step.name not in self.etl_steps_dict():
            self.etl_steps.append(etl_step)
            return
        raise ValueError(
            f"ETLStep named {etl_step.name} already in model please use a new name:\n{self.etl_steps_dict().keys()}"
        )

    def validate_marker(self, marker: str) -> bool:
        """Compare a marker to the models etl_steps to see if it is valid."""
        valid_marker = False
        for etl_step in self.etl_steps:
            if marker == etl_step.name or marker in etl_step.tags:
                valid_marker = True
        return valid_marker

    def _etl_step_graph(self) -> "DiGraph":
        from networkx import DiGraph

        graph = DiGraph()

        # Add edges for every Arg in graph
        edges: List[Tuple[str, str]] = []
        for source_name, source in self.etl_steps_dict().items():
            for target_name, target in self.etl_steps_dict().items():
                if source_name != target_name:
                    if source._get_dependency().test(target._get_dependency()):
                        edges.append((target_name, source_name))

        for etl_step in self.etl_steps_dict().values():
            graph.add_node(etl_step.name, id=etl_step.hash, **etl_step.dict(include={'name', 'dependency'}))
        graph.add_edges_from(edges)
        return graph

    def _sort_graph(self) -> List[ETLStep]:
        graph = self._etl_step_graph()
        etl_steps = self.etl_steps_dict()
        sorted_node_ids = topsort_with_dict(graph)
        sorted_nodes = [etl_steps[key] for key in sorted_node_ids]
        return sorted_nodes

    def run(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        run_config: "RunConfig" = None,
        build: bool = False,
        rerun_failed: bool = False,
        remote: bool = True,
        run_async: bool = True,
    ) -> RunEntity:
        from dbgen.core.run.model_run import ModelRun

        # Check for empty model
        if not self.etl_steps:
            raise ModelRunError(
                f'Model {self.name} has no ETLSteps to run. Did you make sure to instantiate your ETLSteps within the model\'s context?'
            )

        return ModelRun(model=self).execute(
            main_engine=main_engine,
            meta_engine=meta_engine,
            run_config=run_config,
            build=build,
            run_async=run_async,
            remote=remote,
            rerun_failed=rerun_failed,
        )

    def sync(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        build: bool = False,
        meta_only: bool = False,
        create: bool = True,
    ) -> None:
        """Syncs the state of the models registry with the database."""
        if not build and not create:
            raise ValueError("Not building and not creating, sync is doing nothing...")
        if not meta_only:
            if build:
                self.drop_metadata(main_engine, self.registry.metadata)
            if create:
                self.create_metadata(main_engine, self.registry.metadata)
        if build:
            self.drop_metadata(meta_engine, self.meta_registry.metadata)
        if create:
            self.create_metadata(meta_engine, self.meta_registry.metadata)

    def drop_metadata(self, engine: Engine, metadata: MetaData):
        try:
            metadata.drop_all(engine)
        except sqlalchemy.exc.InternalError as exc:
            raise ValueError("Error occurred during rebuilding. Please drop the schema manually...") from exc

    def create_metadata(self, engine: Engine, metadata: MetaData):
        inspector = inspect(engine)
        expected_schemas = {x.schema for x in metadata.tables.values() if x.schema}
        current_schemas = inspector.get_schema_names()
        missing_schema = {x for x in expected_schemas if x not in current_schemas}
        for schema in missing_schema:
            self._logger.debug(f"Missing schema {schema}, creating now")
            with engine.begin() as conn:
                conn.execute(text(f"CREATE SCHEMA {schema}"))
        # Create meta schema
        try:
            metadata.create_all(engine)
        except sqlalchemy.exc.InternalError as exc:
            raise ValueError("Error occurred during database creation") from exc

    def _get_model_row(self):
        graph = self._etl_step_graph()

        def default(obj):
            if isinstance(obj, set):
                return list(obj)
            elif isinstance(obj, UUID):
                return str(obj)
            return obj

        metadata = {'name': self.name, 'id': self.uuid}
        graph_json = serialize_graph(graph, default, metadata=metadata)
        # Assemble stringified dependency fields as we can't store sets in postgres easily
        return ModelEntity(
            id=self.uuid,
            name=self.name,
            etl_steps=[x._get_etl_step_row() for x in self.etl_steps],
            graph_json=loads(graph_json),
        )
